"""Author: Ioannis Pitoskas (jpitoskas@gmail.com)"""

import torch
import numpy as np
import torch.nn as nn
from typing import Optional, Iterable, Union
from torch_geometric.nn import SAGEConv
from torch_geometric.nn import (
    GraphNorm,
)  # , BatchNorm, GraphSizeNorm, InstanceNorm, LayerNorm, Set2Set
from torch_geometric.nn import global_add_pool, global_mean_pool, global_max_pool
import torch.nn.init as init
from torch import Tensor

from .fully_connected_network import FullyConnectedNetwork


class GraphSAGEBlock(nn.Module):
    """A single Graph SAGE Block consisting of a SAGEConv layer, an activation function,
    and a dropout layer. Optionally, a graph normalization layer can be applied.

    Attributes
    ----------
        hidden_layer (torch_geometric.nn.SAGEConv): The SAGEConv layer.
        graph_norm (bool): Whether to apply graph normalization.
        gn_layer (nn.Module or None): The graph normalisation layer.
        activation (nn.Module): Activation function to apply after the hidden layer.
        dropout (nn.Dropout): The dropout layer.
        jittable (bool): Whether to make the hidden module jittable.

    """

    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        activation: nn.Module = nn.ReLU(),
        dropout_probability: float = 0.5,
        graph_norm: bool = False,
        jittable: bool = True,
        *args,
        **kwargs,
    ):
        """Args:
        ----
            input_dim (int): Dimension of the input node features.
            hidden_dim (int): Dimension of the hidden features.
            activation (nn.Module): Activation function to apply after the hidden layer. Default is nn.ReLU().
            dropout_probability (float): Dropout probability. Default is 0.5.
            graph_norm (bool): Whether to apply graph normalization. Default is False.
            jittable (bool): Whether to make the hidden module jittable. Default is True.

        """
        super(GraphSAGEBlock, self).__init__()

        self.jittable = jittable

        self.hidden_layer = SAGEConv(input_dim, hidden_dim)

        if self.jittable:
            self.hidden_layer = self.hidden_layer.jittable()

        self.graph_norm = graph_norm
        if self.graph_norm:
            self.gn_layer = GraphNorm(hidden_dim)
            # self.gn_layer = BatchNorm(hidden_dim)
            # self.gn_layer = GraphSizeNorm()
            # self.gn_layer = InstanceNorm(hidden_dim)
            # self.gn_layer = LayerNorm(hidden_dim)
        else:
            self.gn_layer = None

        self.activation = activation
        self.dropout = nn.Dropout(p=dropout_probability)

    def forward(self, x: Tensor, edge_index: Tensor, batch: Optional[Tensor]) -> Tensor:
        """Forward pass through the layer.

        Args:
        ----
            x (Tensor): Input node features.
            edge_index (Tensor): Graph connectivity in COO format with shape [2, num_edges].
            batch (Optional[Tensor]): Batch vector with shape [num_samples,] which assigns each element to a specific example.

        Returns:
        -------
            Tensor: Output tensor of the layer.

        """
        x = self.hidden_layer(x, edge_index)
        if self.gn_layer is not None:
            x = self.gn_layer(x, batch)
            # x = self.gn_layer(x)
        x = self.activation(x)
        x = self.dropout(x)

        return x


class GraphSAGENetwork(nn.Module):
    """A Graph SAGE Network consisting of multiple Graph SAGE Blocks followed by a Fully Connected Layer.

    Attributes
    ----------
        input_dim (int): Dimension of the input node features.
        hidden_dims (Iterable[int]): Dimensions of the hidden layers.
        output_dim (int): Dimension of the network's output.
        activation (nn.Module): Activation function to apply after each hidden layer.
        dropout_probabilities (list): List of dropout probabilities after each hidden layer.
        graph_norm (bool): Whether to apply graph normalization.
        pooling (str): Type of pooling to apply ('mean', 'add', 'max').
        jittable (bool): Whether to make the GATConv modules jittable.

    """

    def __init__(
        self,
        input_dim: int,
        hidden_dims: Iterable[int],
        output_dim: int = 1,
        activation: nn.Module = nn.ReLU(),
        dropout: Union[float, Iterable[float]] = 0.5,
        graph_norm: bool = False,
        pooling: str = "mean",
        jittable: bool = True,
        *args,
        **kwargs,
    ):
        torch.manual_seed(42)
        np.random.seed(42)
        """
        Args:
            input_dim (int): Dimension of the input node features.
            hidden_dims (Iterable[int]): Dimensions of the hidden layers.
            heads (Union[int, Iterable[int]]): Number of attention heads for each layer. Default is 1.
            edge_dim (Optional[int]): Dimension of the edge features if any. Default is None.
            output_dim (int): Dimension of the network's output. Default is 1.
            activation (nn.Module): Activation function to apply after each hidden layer. Default is nn.ReLU().
            dropout (Union[float, Iterable[float]]): Dropout probability for each layer. Default is 0.5.
            graph_norm (bool): Whether to apply graph normalization. Default is False.
            pooling (str): Type of pooling to apply ('mean', 'add', 'max'). Default is 'mean'.
            jittable (bool): Whether to make the hidden modules jittable. Default is True.
        """

        super(GraphSAGENetwork, self).__init__()

        # Input types check
        if not isinstance(input_dim, int):
            raise TypeError("input_dim must be of type int")

        if not isinstance(hidden_dims, Iterable):
            raise TypeError("hidden_dims must be an Iterable")
        if len(hidden_dims) == 0:
            raise ValueError("hidden_dims must not be empty")
        if not all(isinstance(hidden_dim, int) for hidden_dim in hidden_dims):
            raise TypeError("hidden_dims must only contain integers")

        if not isinstance(output_dim, int):
            raise TypeError("output_dim must be of type int")

        if not isinstance(activation, nn.Module):
            raise TypeError("activation must be a torch.nn.Module")

        if not (isinstance(dropout, float) or isinstance(dropout, Iterable)):
            raise TypeError("dropout must be either of type float or Iterable")
        if isinstance(dropout, float):
            if not 0 <= dropout <= 1:
                raise ValueError("dropout probability must be between 0 and 1")
        if isinstance(dropout, Iterable):
            for item in dropout:
                if not isinstance(item, float):
                    raise TypeError("dropout list must only contain floats")
                if not 0 <= item <= 1:
                    raise ValueError(
                        "Each element in the dropout list must be between 0 and 1"
                    )
            if len(dropout) != len(hidden_dims):
                raise ValueError("hidden_dims and dropout must be of same size")

        if not isinstance(graph_norm, bool):
            raise TypeError("graph_norm must be of type bool")

        if pooling is not None and pooling not in ["mean", "add", "max"]:
            raise NotImplementedError(
                f"Pooling operation '{self.pooling}' is not supported"
            )

        self.input_dim = input_dim
        self.hidden_dims = hidden_dims
        self.output_dim = output_dim
        self.dropout_probabilities = (
            [dropout] * len(hidden_dims) if isinstance(dropout, float) else dropout
        )
        self.graph_norm = graph_norm
        self.pooling = pooling
        self.jittable = jittable

        self.graph_layers = nn.ModuleList()
        graph_layer = GraphSAGEBlock(
            input_dim,
            hidden_dims[0],
            activation=activation,
            dropout_probability=self.dropout_probabilities[0],
            graph_norm=graph_norm,
            jittable=jittable,
        )
        self.graph_layers.append(graph_layer)

        for i in range(len(hidden_dims) - 1):
            graph_layer = GraphSAGEBlock(
                hidden_dims[i],
                hidden_dims[i + 1],
                activation=activation,
                dropout_probability=self.dropout_probabilities[i],
                graph_norm=graph_norm,
                jittable=jittable,
            )
            self.graph_layers.append(graph_layer)

        # self.aggr = Set2Set(hidden_dims[-1], processing_steps=4)
        # Initialise Fully Connected Layer
        self.fc = nn.Linear(hidden_dims[-1], output_dim)

        # Apply Xavier initialization to fc
        init.xavier_uniform_(self.fc.weight)
        init.zeros_(self.fc.bias)

    def forward(self, x: Tensor, edge_index: Tensor, batch: Optional[Tensor]) -> Tensor:
        """Forward pass through the entire network.

        Args:
        ----
            x (Tensor): Input node features.
            edge_index (Tensor): Graph connectivity in COO format with shape [2, num_edges].
            batch (Optional[Tensor]): Batch vector with shape [num_samples,] which assigns each element to a specific example.

        Returns:
        -------
            Tensor: Output tensor after passing through the network.

        """
        x = self._forward_graph(x, edge_index, batch)
        x = self.fc(x)
        return x

    def _forward_graph(
        self, x: Tensor, edge_index: Tensor, batch: Optional[Tensor]
    ) -> Tensor:
        """Helper method for the forward pass through graph layers and pooling."""
        for graph_layer in self.graph_layers:
            x = graph_layer(x, edge_index, batch)
        # x = self.aggr(x, batch)
        x = self._pooling_function(x, batch)
        return x

    def _pooling_function(self, x: Tensor, batch: Optional[Tensor]) -> Tensor:
        """Helper method for the pooling operation.

        Args:
        ----
            x (Tensor): Node feature matrix with shape [num_nodes, ?].
            batch (Optional[Tensor]): The batch vector of shape [num_samples,] which assigns each element to a specific example.

        Returns:
        -------
            Tensor: Output tensor after batch-level aggregation.

        """
        if self.pooling == "add":
            return global_add_pool(x, batch)
        elif self.pooling == "mean":
            return global_mean_pool(x, batch)
        elif self.pooling == "max":
            return global_max_pool(x, batch)
        else:
            raise NotImplementedError(
                f"Pooling operation '{self.pooling}' is not supported"
            )


class GraphSAGENetworkWithExternal(nn.Module):
    """A Graph SAGE Network that integrates external features.
    Combines the output from a Graph SAGE Network with external features
    and processes them through a Fully Connected Network.

    Attributes
    ----------
        graph_model (jaqpotpy.jaqpotpy_torch.models.GraphSAGENetwork): The graph network to treat the graph features.
        num_external_features (int): Number of external features.
        output_dim (int): Dimension of the network's output.
        fc_net (jaqpotpy.jaqpotpy_torch.models.FullyConnectedNetwork): The fc network to treat the external features.

    """

    def __init__(
        self,
        graph_input_dim: int,
        num_external_features: int,
        graph_hidden_dims: Iterable[int],
        fc_hidden_dims: Iterable[int],
        graph_output_dim: int,
        output_dim: int = 1,
        graph_activation: nn.Module = nn.ReLU(),
        graph_dropout: Union[float, Iterable[float]] = 0.5,
        graph_norm: bool = False,
        graph_pooling: str = "mean",
        fc_dropout: Union[float, Iterable[float]] = 0.5,
        fc_activation: nn.Module = nn.ReLU(),
        jittable: bool = True,
        *args,
        **kwargs,
    ):
        """Args:
        ----
            graph_input_dim (int): Dimension of the input node features for the graph.
            num_external_features (int): Number of external features.
            graph_hidden_dims (Iterable[int]): Dimensions of the hidden layers in the graph.
            fc_hidden_dims (Iterable[int]): Dimensions of the hidden layers in the fully connected network.
            graph_output_dim (int): Dimension of the graph output features.
            output_dim (int): Dimension of the network's output. Default is 1.
            graph_activation (nn.Module): Activation function for graph layers. Default is nn.ReLU().
            graph_dropout (Union[float, Iterable[float]]): Dropout probability for graph layers. Default is 0.5
            graph_norm (bool): Whether to apply graph normalization in the graph layers. Default is False.
            graph_pooling (str): Type of pooling to apply in the graph layers ('mean', 'add', 'max'). Default is 'mean'.
            fc_dropout (Union[float, Iterable[float]]): Dropout probability for fully connected layers. Default is 0.5.
            fc_activation (nn.Module): Activation function for fully connected layers. Default is nn.ReLU().
            jittable (bool): Whether to make the hidden modules jittable. Default is True.

        """
        super().__init__()

        if not isinstance(num_external_features, int):
            raise TypeError("num_external_features must be of type int")

        self.graph_model = GraphSAGENetwork(
            input_dim=graph_input_dim,
            hidden_dims=graph_hidden_dims,
            output_dim=graph_output_dim,
            activation=graph_activation,
            dropout=graph_dropout,
            graph_norm=graph_norm,
            pooling=graph_pooling,
            jittable=jittable,
        )

        self.num_external_features = num_external_features
        self.output_dim = output_dim

        self.fc_net = FullyConnectedNetwork(
            input_dim=graph_output_dim + self.num_external_features,
            hidden_dims=fc_hidden_dims,
            output_dim=self.output_dim,
            activation=fc_activation,
            dropout=fc_dropout,
        )

    def forward(
        self, x: Tensor, edge_index: Tensor, external: Tensor, batch: Optional[Tensor]
    ) -> Tensor:
        """Forward pass through the entire network.

        Args:
        ----
            x (Tensor): Node feature matrix with shape [num_nodes, num_node_features].
            edge_index (Tensor): Graph connectivity in COO format with shape [2, num_edges].
            external (Tensor): External feature matrix with shape [num_samples, num_external_features].
            batch (Optional[Tensor]): The batch vector of shape [num_samples,] which assigns each element to a specific example.
            edge_attr (OptTensor): Edge feature matrix with shape [num_edges, num_edge_features]. Default is None.

        Returns:
        -------
            Tensor: Output tensor after passing through the network.

        """
        x = self.graph_model(x, edge_index, batch)
        x = torch.cat((x, external), dim=1)
        x = self.fc_net(x)
        return x



import inspect
import logging
import numpy as np
import pandas as pd
from typing import Any, Union, List, Generator
import os
from jaqpotpy.entities.material_models import Atoms

logger = logging.getLogger(__name__)
_print_threshold = 10


class Parser(object):
    """Abstract class for parsing molecules or materials files into serializable objects.
    This class is abstract and cannot be invoked directly. You'll
    likely only interact with this class if you're a developer. In
    that case, you might want to make a child class which
    implements the `_parse` (or the `_parse_dataframe`) method for calculating parsing for
    parsing files representing materials or molecules.
    """

    def __init__(self, path: str, file_ext: Union[str, List[str]]):
        self.path = path
        if self.path[-1] == "/":
            self.path = self.path[:-1]

        self.file_ext = file_ext
        if isinstance(self.file_ext, str):
            self.file_ext = [self.file_ext]

        self.files_ = []

    def parse(self) -> Generator:
        """Parse files representing materials or molecules.

        Parameters
        ----------
        path: str
          Either the path of a certain file or a path of a folder containing
          files that will be parsed

        file_ext: str | List[str]
          The extention of the files that will be parsed.

        log_every_n: int, default 1000
          Logs featurization progress every `log_every_n` steps.

        Returns
        -------
        generator
          A generator with the parsed objects of the files.

        """
        try:
            if self.path.split(".")[-1].lower() not in self.file_ext:
                try:
                    file = [
                        self.path + "/" + item
                        for item in os.listdir(self.path)
                        if item.split(".")[-1].lower() in self.file_ext
                    ]
                except:
                    raise ValueError(
                        "Invalid file type. Expected {} but found {} instead".format(
                            self.file_ext, self.path.split(".")[-1]
                        )
                    )
            else:
                file = [self.path]
        except:
            try:
                file = [
                    self.path + "/" + item
                    for item in os.listdir(self.path)
                    if item.split(".")[-1].lower() in self.file_ext
                ]
            except:
                raise ValueError(
                    "Invalid file type. Expected {} but didn't find any".format(
                        self.file_ext
                    )
                )

        # Open the file and read it in as a list of rows
        for i, item in enumerate(file):
            try:
                obj = self._parse(item)
                if isinstance(obj, list):
                    for i in obj:
                        yield i
                else:
                    yield obj
            except Exception as e:
                logger.warning("Failed to parse file {}.".format(item), e)

    def parse_dataframe(self) -> Union[List[pd.DataFrame], pd.DataFrame]:
        """Parse files of materials or molecules in a pandas dataframe.

        Parameters
        ----------
        path: str
          Either the path of a certain file or a path of a folder containing
          files that will be parsed

        file_ext: str | List[str]
          The extention of the files that will be parsed.

        log_every_n: int, default 1000
          Logs featurization progress every `log_every_n` steps.

        Returns
        -------
        pd.DataFrame
          A pandas dataframe of the files.

        """
        files = self.parse()

        df = pd.DataFrame()
        df_list = []
        for i, file in enumerate(files):
            resp = self._parse_dataframe(file, self.files_[i])
            try:
                df = pd.concat([df, resp]).reset_index(drop=True)
            except:
                if len(df_list) == 0:
                    df_list = [
                        pd.concat([df, item]).reset_index(drop=True) for item in resp
                    ]
                else:
                    for j in range(len(df_list)):
                        df_list[j] = pd.concat([df_list[j], resp[j]]).reset_index(
                            drop=True
                        )

        if len(df_list) > 0:
            return df_list
        return df

    def __call__(self, path: str, file_ext: Union[str, List[str]], **kwargs):
        """Calculate features for datapoints.
        `**kwargs` will get passed directly to `Featurizer.featurize`

        Parameters
        ----------
        datapoints: Iterable[Any]
          Any blob of data you like. Subclasss should instantiate this.

        """
        return self.parse()

    def _parse(self, path: str):
        """Calculate features for a single datapoint.

        Parameters
        ----------
        datapoint: Any
          Any blob of data you like. Subclass should instantiate this.

        """
        raise NotImplementedError("Parser is not defined.")

    def _parse_dataframe(
        self, file: Any, filename: str
    ) -> Union[List[pd.DataFrame], pd.DataFrame]:
        """Calculate features for a single datapoint.

        Parameters
        ----------
        datapoint: Any
          Any blob of data you like. Subclass should instantiate this.

        """
        raise NotImplementedError("Parser is not defined.")

    def __repr__(self) -> str:
        """Convert self to repr representation.

        Returns
        -------
        str
          The string represents the class.

        Examples
        --------
        >>> import jaqpotpy as jt
        >>> jt.parsers.PdbParser('pdb_file.pdb')
        PdbParser[pdb='pdb_file.pdb', file_ext='pdb']

        """
        args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        args_names = [arg for arg in args_spec.args if arg != "self"]
        args_info = ""
        for arg_name in args_names:
            value = self.__dict__[arg_name]
            # for str
            if isinstance(value, str):
                value = "'" + value + "'"
            # for list
            if isinstance(value, list):
                threshold = get_print_threshold()
                value = np.array2string(np.array(value), threshold=threshold)
            args_info += arg_name + "=" + str(value) + ", "
        return self.__class__.__name__ + "[" + args_info[:-2] + "]"

    def __str__(self) -> str:
        """Convert self to str representation.

        Returns
        -------
        str
          The string represents the class.

        Examples
        --------
        >>> import jaqpotpy as jt
        >>> jt.parsers.PdbParser('pdb_file.pdb')
        PdbParser[pdb='pdb_file.pdb', file_ext='pdb']

        """
        args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        args_names = [arg for arg in args_spec.args if arg != "self"]
        args_num = len(args_names)
        args_default_values = [None for _ in range(args_num)]
        if args_spec.defaults is not None:
            defaults = list(args_spec.defaults)
            args_default_values[-len(defaults) :] = defaults

        override_args_info = ""
        for arg_name, default in zip(args_names, args_default_values):
            if arg_name in self.__dict__:
                arg_value = self.__dict__[arg_name]
                # validation
                # skip list
                if isinstance(arg_value, list):
                    continue
                if isinstance(arg_value, str):
                    # skip path string
                    if "\\/." in arg_value or "/" in arg_value or "." in arg_value:
                        continue
                # main logic
                if default != arg_value:
                    override_args_info += "_" + arg_name + "_" + str(arg_value)
        return self.__class__.__name__ + override_args_info

    def to_xyz(self):
        xyzs = []
        files = self.parse()
        for file in files:
            atoms: Atoms = file.get_atoms()
            xyz = str(len(atoms.elements)) + "\ntemp xyz\n"
            for i in range(len(atoms.elements)):
                xyz += (
                    atoms.elements[i]
                    + "\t"
                    + str(atoms.coordinates[i][0])
                    + "\t"
                    + str(atoms.coordinates[i][1])
                    + "\t"
                    + str(atoms.coordinates[i][2])
                    + "\n"
                )

            xyzs.append(xyz)

        return xyzs


def get_print_threshold() -> int:
    """Return the printing threshold for datasets.
    The print threshold is the number of elements from ids/tasks to
    print when printing representations of `Dataset` objects.

    Returns
    -------
    threshold: int
      Number of elements that will be printed

    """
    return _print_threshold



from typing import List
from jaqpotpy.parsers.base_classes import Parser
from jaqpotpy.entities.material_models import (
    Atoms,
    Sdf,
    SdfHeader,
    SdfCTable,
    SdfBond,
    SdfCountLine,
)
import pandas as pd
from pydantic import parse_obj_as


def _clean_key(word) -> str:
    return "".join(i for i in word if i.isalnum())


class MolParser(Parser):
    """MolParser.
    This class parses a mol or sdf file (or all mol or sdf files in a folder) into a Sdf structure.
    For more information on the structure see jaqpotpy.models.material_models.Sdf

    Attributes:
    ----------
    files_: str | List[str]
        File names of the pdb files that were parsed.

    Note:
    ----

    Examples:
    --------
    >>> import jaqpotpy as jt
    >>> mol_file = './caffeine.mol'
    >>> parser = jt.parsers.mol_parser.mol_parser.MolParser()
    >>> mol = parser.parse(mol_file)
    >>> type(mol)
    generator
    >>> parsed = next(mol)
    >>> type(parsed)
    jaqpotpy.models.material_models.Sdf

    """

    @property
    def __name__(self):
        return "MolParser"

    def __getitem__(self):
        return self

    def _parse(self, path: str) -> Sdf:
        """Parse sdf and mol files.

        Parameters
        ----------
        path: str
          Either the path of a certain file or a path of a folder containing
          files that will be parsed

        Returns
        -------
        jaqpotpy.models.material_models.Sdf
          A Sdf object

        """
        # Initialize variables
        sdf_dict: Sdf = Sdf(
            header=SdfHeader(title="", timestamp="", comment=""),
            con_table=SdfCTable(
                count_line=SdfCountLine(
                    n_atoms=0,
                    n_bonds=0,
                    n_atom_list=0,
                    chiral_flag=0,
                    n_stext_ents=0,
                    n_add_props=0,
                    version="",
                ),
                atoms=Atoms(elements=[], coordinates=[], extraInfo=[]),
                bonds=[
                    SdfBond(
                        first_atom=0,
                        second_atom=0,
                        bond_type=0,
                        bond_stereo=0,
                        bond_topology=0,
                        reacting_center_status=0,
                    )
                ],
                properties=[],
            ),
            extraInfo=[],
        )

        head: SdfHeader = SdfHeader(title="", timestamp="", comment="")
        ctable: SdfCTable = SdfCTable(
            count_line=SdfCountLine(
                n_atoms=0,
                n_bonds=0,
                n_atom_list=0,
                chiral_flag=0,
                n_stext_ents=0,
                n_add_props=0,
                version="",
            ),
            atoms=Atoms(elements=[], coordinates=[], extraInfo=[]),
            bonds=[
                SdfBond(
                    first_atom=0,
                    second_atom=0,
                    bond_type=0,
                    bond_stereo=0,
                    bond_topology=0,
                    reacting_center_status=0,
                )
            ],
            properties=[],
        )
        records = []

        # Open the file and read it in as a list of rows
        with open(path) as f:
            sdf = f.read().splitlines()

        # Iterate through the file
        line_cnt = 0
        extraProp = ""
        cont = True

        for row in sdf:
            if line_cnt == 0:
                head.title = row
            elif line_cnt == 1:
                head.timestamp = row
            elif line_cnt == 2:
                head.comment = row
            elif line_cnt == 3:
                curr_list = row.split()
                try:
                    ctable.count_line.n_atoms = int(curr_list[0])
                    ctable.count_line.n_bonds = int(curr_list[1])
                    ctable.count_line.n_atom_list = int(curr_list[2])
                    ctable.count_line.chiral_flag = int(curr_list[3])
                    ctable.count_line.n_stext_ents = int(curr_list[4])
                    ctable.count_line.n_add_props = int(curr_list[5])
                    ctable.count_line.version = curr_list[6]
                except:
                    raise TypeError("Count line has wrong format. Count Line: ", row)
            else:
                curr_list = row.split()
                try:
                    if cont:
                        _ = int(curr_list[0])
                    else:
                        _ = int("")
                except ValueError:
                    if row.strip() != "":
                        if extraProp != "":
                            sdf_dict.extraInfo.append(extraProp + row)
                            extraProp = ""
                        elif curr_list[0].strip().lower() == "m":
                            if curr_list[1].strip().lower() != "end":
                                ctable.properties.append(row)
                            else:
                                cont = False
                                sdf_dict.header = head
                                sdf_dict.con_table = ctable
                        elif curr_list[0].strip() == ">":
                            if extraProp == "":
                                extraProp += (
                                    row[1:][row[1:].find("<") + 1 : row[1:].find(">")]
                                    + ": "
                                )

                        elif row[0] == "$":
                            records.append(sdf_dict)
                            self.files_.append(path)
                            sdf_dict: Sdf = Sdf(
                                header=SdfHeader(title="", timestamp="", comment=""),
                                con_table=SdfCTable(
                                    count_line=SdfCountLine(
                                        n_atoms=0,
                                        n_bonds=0,
                                        n_atom_list=0,
                                        chiral_flag=0,
                                        n_stext_ents=0,
                                        n_add_props=0,
                                        version="",
                                    ),
                                    atoms=Atoms(
                                        elements=[], coordinates=[], extraInfo=[]
                                    ),
                                    bonds=[
                                        SdfBond(
                                            first_atom=0,
                                            second_atom=0,
                                            bond_type=0,
                                            bond_stereo=0,
                                            bond_topology=0,
                                            reacting_center_status=0,
                                        )
                                    ],
                                    properties=[],
                                ),
                                extraInfo=[],
                            )

                            head: SdfHeader = SdfHeader(
                                title="", timestamp="", comment=""
                            )
                            ctable: SdfCTable = SdfCTable(
                                count_line=SdfCountLine(
                                    n_atoms=0,
                                    n_bonds=0,
                                    n_atom_list=0,
                                    chiral_flag=0,
                                    n_stext_ents=0,
                                    n_add_props=0,
                                    version="",
                                ),
                                atoms=Atoms(elements=[], coordinates=[], extraInfo=[]),
                                bonds=[
                                    SdfBond(
                                        first_atom=0,
                                        second_atom=0,
                                        bond_type=0,
                                        bond_stereo=0,
                                        bond_topology=0,
                                        reacting_center_status=0,
                                    )
                                ],
                                properties=[],
                            )
                            line_cnt = -1
                        else:
                            try:
                                ctable.atoms.elements.append(_clean_key(curr_list[3]))
                                ctable.atoms.coordinates.append(
                                    [
                                        float(curr_list[0]),
                                        float(curr_list[1]),
                                        float(curr_list[2]),
                                    ]
                                )
                                ctable.atoms.extraInfo.append(curr_list[4:])
                            except:
                                # empty row
                                pass
                except IndexError:
                    pass
                except Exception as e:
                    raise TypeError("File has malformed fields. Error: ", str(e))
                else:
                    try:
                        ctable.bonds.append(
                            parse_obj_as(
                                SdfBond,
                                {
                                    "first_atom": int(curr_list[0]),
                                    "second_atom": int(curr_list[1]),
                                    "bond_type": int(curr_list[2]),
                                    "bond_stereo": int(curr_list[3]),
                                    "bond_topology": int(curr_list[4]),
                                    "reacting_center_status": int(curr_list[5]),
                                },
                            )
                        )
                    except:
                        raise TypeError("Found bond with wrong format. Bond: ", row)

            line_cnt += 1

        if len(records) > 1:
            return records

        try:
            return records[0]
        except:
            return sdf_dict

    def _parse_dataframe(self, file: Sdf, filename: str) -> List[pd.DataFrame]:
        """Parse mol or sdf files in a pandas dataframe.

        Parameters
        ----------
        file: jaqpotpy.models.material_models.Sdf
            The Sdf structrure of a parsed file

        filename: str
            The name of the parsed file

        Returns
        -------
        List[pd.DataFrame()]

        """
        atoms = pd.DataFrame()
        bonds = pd.DataFrame()

        for i in range(len(file.con_table.atoms.elements)):
            d = {}
            d["file"] = filename
            d["element"] = file.con_table.atoms.elements[i]
            d["x"] = file.con_table.atoms.coordinates[i][0]
            d["y"] = file.con_table.atoms.coordinates[i][1]
            d["z"] = file.con_table.atoms.coordinates[i][2]

            atoms = pd.concat([atoms, pd.DataFrame(d, index=[0])]).reset_index(
                drop=True
            )

        for bond in file.con_table.bonds:
            d = bond.dict()
            d["file"] = filename
            bonds = pd.concat([bonds, pd.DataFrame(d, index=[0])]).reset_index(
                drop=True
            )

        return [atoms, bonds]



"""This module adds utilities for coordinate boxes"""

import os
from typing import List, Sequence, Tuple
from typing import Optional

import numpy as np
from scipy.spatial import ConvexHull

from jaqpotpy.utils.pdbqt_utils import pdbqt_to_pdb
from jaqpotpy.utils.rdkit_utils import load_molecule, write_molecule
from jaqpotpy.utils.types import RDKitMol

"""This module adds utilities for coordinate boxes"""


def create_hydrated_pdbqt_pdb(protein_file: str, out_dir: str = "./"):
    protein_name = os.path.basename(protein_file).split(".")[0]
    protein_hyd = os.path.join(out_dir, "%s_hyd.pdb" % protein_name)
    protein_pdbqt = os.path.join(out_dir, "%s.pdbqt" % protein_name)
    protein_mol = load_molecule(protein_file, calc_charges=True, add_hydrogens=True)
    write_molecule(protein_mol[1], protein_hyd, is_protein=True)
    write_molecule(protein_mol[1], protein_pdbqt, is_protein=True)
    return protein_pdbqt


class CoordinateBox(object):
    """A coordinate box that represents a block in space.
    Molecular complexes are typically represented with atoms as
    coordinate points. Each complex is naturally associated with a
    number of different box regions. For example, the bounding box is a
    box that contains all atoms in the molecular complex. A binding
    pocket box is a box that focuses in on a binding region of a protein
    to a ligand. A interface box is the region in which two proteins
    have a bulk interaction.
    The `CoordinateBox` class is designed to represent such regions of
    space. It consists of the coordinates of the box, and the collection
    of atoms that live in this box alongside their coordinates.
    """

    def __init__(
            self,
            x_range: Tuple[float, float],
            y_range: Tuple[float, float],
            z_range: Tuple[float, float],
    ):
        """Initialize this box.

        Parameters
        ----------
        x_range: Tuple[float, float]
          A tuple of `(x_min, x_max)` with max and min x-coordinates.
        y_range: Tuple[float, float]
          A tuple of `(y_min, y_max)` with max and min y-coordinates.
        z_range: Tuple[float, float]
          A tuple of `(z_min, z_max)` with max and min z-coordinates.

        Raises
        ------
        `ValueError` if this interval is malformed

        """
        if not isinstance(x_range, tuple) or not len(x_range) == 2:
            raise ValueError("x_range must be a tuple of length 2")
        else:
            x_min, x_max = x_range
            if not x_min <= x_max:
                raise ValueError("x minimum must be <= x maximum")
        if not isinstance(y_range, tuple) or not len(y_range) == 2:
            raise ValueError("y_range must be a tuple of length 2")
        else:
            y_min, y_max = y_range
            if not y_min <= y_max:
                raise ValueError("y minimum must be <= y maximum")
        if not isinstance(z_range, tuple) or not len(z_range) == 2:
            raise ValueError("z_range must be a tuple of length 2")
        else:
            z_min, z_max = z_range
            if not z_min <= z_max:
                raise ValueError("z minimum must be <= z maximum")
        self.x_range = x_range
        self.y_range = y_range
        self.z_range = z_range

    def __repr__(self):
        """Create a string representation of this box"""
        x_str = str(self.x_range)
        y_str = str(self.y_range)
        z_str = str(self.z_range)
        return "Box[x_bounds=%s, y_bounds=%s, z_bounds=%s]" % (x_str, y_str, z_str)

    def __str__(self):
        """Create a string representation of this box."""
        return self.__repr__()

    def __contains__(self, point: Sequence[float]) -> bool:
        """Check whether a point is in this box.

        Parameters
        ----------
        point: Sequence[float]
          3-tuple or list of length 3 or np.ndarray of shape `(3,)`.
          The `(x, y, z)` coordinates of a point in space.

        Returns
        -------
        bool
          `True` if `other` is contained in this box.

        """
        (x_min, x_max) = self.x_range
        (y_min, y_max) = self.y_range
        (z_min, z_max) = self.z_range
        x_cont = x_min <= point[0] and point[0] <= x_max
        y_cont = y_min <= point[1] and point[1] <= y_max
        z_cont = z_min <= point[2] and point[2] <= z_max
        return x_cont and y_cont and z_cont

    # FIXME: Argument 1 of "__eq__" is incompatible with supertype "object"
    def __eq__(self, other: "CoordinateBox") -> bool:  # type: ignore
        """Compare two boxes to see if they're equal.

        Parameters
        ----------
        other: CoordinateBox
          Compare this coordinate box to the other one.

        Returns
        -------
        bool
          That's `True` if all bounds match.

        Raises
        ------
        `ValueError` if attempting to compare to something that isn't a
        `CoordinateBox`.

        """
        if not isinstance(other, CoordinateBox):
            raise ValueError("Can only compare to another box.")
        return (
                self.x_range == other.x_range
                and self.y_range == other.y_range
                and self.z_range == other.z_range
        )

    def __hash__(self) -> int:
        """Implement hashing function for this box.
        Uses the default `hash` on `self.x_range, self.y_range,
        self.z_range`.

        Returns
        -------
        int
          Unique integer

        """
        return hash((self.x_range, self.y_range, self.z_range))

    def center(self) -> Tuple[float, float, float]:
        """Computes the center of this box.

        Returns
        -------
        Tuple[float, float, float]
          `(x, y, z)` the coordinates of the center of the box.

        Examples
        --------
        >>> box = CoordinateBox((0, 1), (0, 1), (0, 1))
        >>> box.center()
        (0.5, 0.5, 0.5)

        """
        x_min, x_max = self.x_range
        y_min, y_max = self.y_range
        z_min, z_max = self.z_range
        return (
            x_min + (x_max - x_min) / 2,
            y_min + (y_max - y_min) / 2,
            z_min + (z_max - z_min) / 2,
        )

    def volume(self) -> float:
        """Computes and returns the volume of this box.

        Returns
        -------
        float
          The volume of this box. Can be 0 if box is empty

        Examples
        --------
        >>> box = CoordinateBox((0, 1), (0, 1), (0, 1))
        >>> box.volume()
        1

        """
        x_min, x_max = self.x_range
        y_min, y_max = self.y_range
        z_min, z_max = self.z_range
        return (x_max - x_min) * (y_max - y_min) * (z_max - z_min)

    def contains(self, other: "CoordinateBox") -> bool:
        """Test whether this box contains another.
        This method checks whether `other` is contained in this box.

        Parameters
        ----------
        other: CoordinateBox
          The box to check is contained in this box.

        Returns
        -------
        bool
          `True` if `other` is contained in this box.

        Raises
        ------
        `ValueError` if `not isinstance(other, CoordinateBox)`.

        """
        if not isinstance(other, CoordinateBox):
            raise ValueError("other must be a CoordinateBox")
        other_x_min, other_x_max = other.x_range
        other_y_min, other_y_max = other.y_range
        other_z_min, other_z_max = other.z_range
        self_x_min, self_x_max = self.x_range
        self_y_min, self_y_max = self.y_range
        self_z_min, self_z_max = self.z_range
        return (
                self_x_min <= other_x_min
                and other_x_max <= self_x_max
                and self_y_min <= other_y_min
                and other_y_max <= self_y_max
                and self_z_min <= other_z_min
                and other_z_max <= self_z_max
        )


def intersect_interval(
        interval1: Tuple[float, float], interval2: Tuple[float, float]
) -> Tuple[float, float]:
    """Computes the intersection of two intervals.

    Parameters
    ----------
    interval1: Tuple[float, float]
      Should be `(x1_min, x1_max)`
    interval2: Tuple[float, float]
      Should be `(x2_min, x2_max)`

    Returns
    -------
    x_intersect: Tuple[float, float]
      Should be the intersection. If the intersection is empty returns
      `(0, 0)` to represent the empty set. Otherwise is `(max(x1_min,
      x2_min), min(x1_max, x2_max))`.

    """
    x1_min, x1_max = interval1
    x2_min, x2_max = interval2
    if x1_max < x2_min:
        # If interval1 < interval2 entirely
        return (0, 0)
    elif x2_max < x1_min:
        # If interval2 < interval1 entirely
        return (0, 0)
    x_min = max(x1_min, x2_min)
    x_max = min(x1_max, x2_max)
    return (x_min, x_max)


def intersection(box1: CoordinateBox, box2: CoordinateBox) -> CoordinateBox:
    """Computes the intersection box of provided boxes.

    Parameters
    ----------
    box1: CoordinateBox
      First `CoordinateBox`
    box2: CoordinateBox
      Another `CoordinateBox` to intersect first one with.

    Returns
    -------
    CoordinateBox
      A `CoordinateBox` containing the intersection. If the intersection is empty,
      returns the box with 0 bounds.

    """
    x_intersection = intersect_interval(box1.x_range, box2.x_range)
    y_intersection = intersect_interval(box1.y_range, box2.y_range)
    z_intersection = intersect_interval(box1.z_range, box2.z_range)
    return CoordinateBox(x_intersection, y_intersection, z_intersection)


def union(box1: CoordinateBox, box2: CoordinateBox) -> CoordinateBox:
    """Merges provided boxes to find the smallest union box.
    This method merges the two provided boxes.

    Parameters
    ----------
    box1: CoordinateBox
      First box to merge in
    box2: CoordinateBox
      Second box to merge into this box

    Returns
    -------
    CoordinateBox
      Smallest `CoordinateBox` that contains both `box1` and `box2`

    """
    x_min = min(box1.x_range[0], box2.x_range[0])
    y_min = min(box1.y_range[0], box2.y_range[0])
    z_min = min(box1.z_range[0], box2.z_range[0])
    x_max = max(box1.x_range[1], box2.x_range[1])
    y_max = max(box1.y_range[1], box2.y_range[1])
    z_max = max(box1.z_range[1], box2.z_range[1])
    return CoordinateBox((x_min, x_max), (y_min, y_max), (z_min, z_max))


def merge_overlapping_boxes(
        boxes: List[CoordinateBox], threshold: float = 0.8
) -> List[CoordinateBox]:
    """Merge boxes which have an overlap greater than threshold.

    Parameters
    ----------
    boxes: list[CoordinateBox]
      A list of `CoordinateBox` objects.
    threshold: float, default 0.8
      The volume fraction of the boxes that must overlap for them to be
      merged together.

    Returns
    -------
    List[CoordinateBox]
      List[CoordinateBox] of merged boxes. This list will have length less
      than or equal to the length of `boxes`.

    """
    outputs: List[CoordinateBox] = []
    for box in boxes:
        for other in boxes:
            if box == other:
                continue
            intersect_box = intersection(box, other)
            if (
                    intersect_box.volume() >= threshold * box.volume()
                    or intersect_box.volume() >= threshold * other.volume()
            ):
                box = union(box, other)
        unique_box = True
        for output in outputs:
            if output.contains(box):
                unique_box = False
        if unique_box:
            outputs.append(box)
    return outputs


def get_face_boxes(coords: np.ndarray, pad: float = 5.0) -> List[CoordinateBox]:
    """For each face of the convex hull, compute a coordinate box around it.
    The convex hull of a macromolecule will have a series of triangular
    faces. For each such triangular face, we construct a bounding box
    around this triangle. Think of this box as attempting to capture
    some binding interaction region whose exterior is controlled by the
    box. Note that this box will likely be a crude approximation, but
    the advantage of this technique is that it only uses simple geometry
    to provide some basic biological insight into the molecule at hand.
    The `pad` parameter is used to control the amount of padding around
    the face to be used for the coordinate box.

    Parameters
    ----------
    coords: np.ndarray
      A numpy array of shape `(N, 3)`. The coordinates of a molecule.
    pad: float, optional (default 5.0)
      The number of angstroms to pad.

    Returns
    -------
    boxes: List[CoordinateBox]
      List of `CoordinateBox`

    Examples
    --------
    >>> coords = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]])
    >>> boxes = get_face_boxes(coords, pad=5)

    """
    hull = ConvexHull(coords)
    boxes = []
    # Each triangle in the simplices is a set of 3 atoms from
    # coordinates which forms the vertices of an exterior triangle on
    # the convex hull of the macromolecule.
    for triangle in hull.simplices:
        # Points is the set of atom coordinates that make up this
        # triangular face on the convex hull
        points = np.array(
            [coords[triangle[0]], coords[triangle[1]], coords[triangle[2]]]
        )
        # Let's extract x/y/z coords for this face
        x_coords = points[:, 0]
        y_coords = points[:, 1]
        z_coords = points[:, 2]

        # Let's compute min/max points
        x_min, x_max = np.amin(x_coords), np.amax(x_coords)
        x_min, x_max = int(np.floor(x_min)) - pad, int(np.ceil(x_max)) + pad
        x_bounds = (x_min, x_max)

        y_min, y_max = np.amin(y_coords), np.amax(y_coords)
        y_min, y_max = int(np.floor(y_min)) - pad, int(np.ceil(y_max)) + pad
        y_bounds = (y_min, y_max)
        z_min, z_max = np.amin(z_coords), np.amax(z_coords)
        z_min, z_max = int(np.floor(z_min)) - pad, int(np.ceil(z_max)) + pad
        z_bounds = (z_min, z_max)
        box = CoordinateBox(x_bounds, y_bounds, z_bounds)
        boxes.append(box)
    return boxes


def write_vina_conf(
        protein_filename: str,
        ligand_filename: str,
        centroid: np.ndarray,
        box_dims: np.ndarray,
        conf_filename: str,
        num_modes: int = 9,
        exhaustiveness: int = None,
) -> None:
    """Writes Vina configuration file to disk.
    Autodock Vina accepts a configuration file which provides options
    under which Vina is invoked. This utility function writes a vina
    configuration file which directs Autodock vina to perform docking
    under the provided options.

    Parameters
    ----------
    protein_filename: str
      Filename for protein
    ligand_filename: str
      Filename for the ligand
    centroid: np.ndarray
      A numpy array with shape `(3,)` holding centroid of system
    box_dims: np.ndarray
      A numpy array of shape `(3,)` holding the size of the box to dock
    conf_filename: str
      Filename to write Autodock Vina configuration to.
    num_modes: int, optional (default 9)
      The number of binding modes Autodock Vina should find
    exhaustiveness: int, optional
      The exhaustiveness of the search to be performed by Vina

    """
    with open(conf_filename, "w") as f:
        f.write("receptor = %s\n" % protein_filename)
        f.write("ligand = %s\n\n" % ligand_filename)

        f.write("center_x = %f\n" % centroid[0])
        f.write("center_y = %f\n" % centroid[1])
        f.write("center_z = %f\n\n" % centroid[2])

        f.write("size_x = %f\n" % box_dims[0])
        f.write("size_y = %f\n" % box_dims[1])
        f.write("size_z = %f\n\n" % box_dims[2])

        f.write("num_modes = %d\n\n" % num_modes)
        if exhaustiveness is not None:
            f.write("exhaustiveness = %d\n" % exhaustiveness)


def write_gnina_conf(
        protein_filename: str,
        ligand_filename: str,
        conf_filename: str,
        num_modes: int = 9,
        exhaustiveness: int = None,
        **kwargs,
) -> None:
    """Writes GNINA configuration file to disk.
    GNINA accepts a configuration file which provides options
    under which GNINA is invoked. This utility function writes a
    configuration file which directs GNINA to perform docking
    under the provided options.

    Parameters
    ----------
    protein_filename: str
      Filename for protein
    ligand_filename: str
      Filename for the ligand
    conf_filename: str
      Filename to write Autodock Vina configuration to.
    num_modes: int, optional (default 9)
      The number of binding modes GNINA should find
    exhaustiveness: int, optional
      The exhaustiveness of the search to be performed by GNINA
    kwargs:
      Args supported by GNINA documented here
      https://github.com/gnina/gnina#usage

    """
    with open(conf_filename, "w") as f:
        f.write("receptor = %s\n" % protein_filename)
        f.write("ligand = %s\n\n" % ligand_filename)

        f.write("autobox_ligand = %s\n\n" % protein_filename)

        if exhaustiveness is not None:
            f.write("exhaustiveness = %d\n" % exhaustiveness)
        f.write("num_modes = %d\n\n" % num_modes)

        for k, v in kwargs.items():
            f.write("%s = %s\n" % (str(k), str(v)))


def read_gnina_log(log_file: str) -> np.ndarray:
    """Read GNINA logfile and get docking scores.
    GNINA writes computed binding affinities to a logfile.

    Parameters
    ----------
    log_file: str
      Filename of logfile generated by GNINA.

    Returns
    -------
    scores: np.array, dimension (num_modes, 3)
      Array of binding affinity (kcal/mol), CNN pose score,
      and CNN affinity for each binding mode.

    """
    scores = []
    lines = open(log_file).readlines()
    mode_start = np.inf
    for idx, line in enumerate(lines):
        if line[:6] == "-----+":
            mode_start = idx
        if idx > mode_start:
            mode = line.split()
            score = [float(x) for x in mode[1:]]
            scores.append(score)

    return np.array(scores)


def load_docked_ligands(pdbqt_output: str) -> Tuple[List[RDKitMol], List[float]]:
    """This function loads ligands docked by autodock vina.
    Autodock vina writes outputs to disk in a PDBQT file format. This
    PDBQT file can contain multiple docked "poses". Recall that a pose
    is an energetically favorable 3D conformation of a molecule. This
    utility function reads and loads the structures for multiple poses
    from vina's output file.

    Parameters
    ----------
    pdbqt_output: str
      Should be the filename of a file generated by autodock vina's
      docking software.

    Returns
    -------
    Tuple[List[rdkit.Chem.rdchem.Mol], List[float]]
      Tuple of `molecules, scores`. `molecules` is a list of rdkit
      molecules with 3D information. `scores` is the associated vina
      score.

    Notes
    -----
    This function requires RDKit to be installed.

    """
    try:
        from rdkit import Chem
    except ModuleNotFoundError:
        raise ImportError("This function requires RDKit to be installed.")

    lines = open(pdbqt_output).readlines()
    molecule_pdbqts = []
    scores = []
    current_pdbqt: Optional[List[str]] = None
    for line in lines:
        if line[:5] == "MODEL":
            current_pdbqt = []
        elif line[:19] == "REMARK VINA RESULT:":
            words = line.split()
            # the line has format
            # REMARK VINA RESULT: score ...
            # There is only 1 such line per model so we can append it
            scores.append(float(words[3]))
        elif line[:6] == "ENDMDL":
            molecule_pdbqts.append(current_pdbqt)
            current_pdbqt = None
        else:
            # FIXME: Item "None" of "Optional[List[str]]" has no attribute "append"
            current_pdbqt.append(line)  # type: ignore

    molecules = []
    for pdbqt_data in molecule_pdbqts:
        pdb_block = pdbqt_to_pdb(pdbqt_data=pdbqt_data)
        mol = Chem.MolFromPDBBlock(str(pdb_block), sanitize=False, removeHs=False)
        molecules.append(mol)
    return molecules, scores


def prepare_inputs(
        protein: str,
        ligand: str,
        replace_nonstandard_residues: bool = True,
        remove_heterogens: bool = True,
        remove_water: bool = True,
        add_hydrogens: bool = True,
        pH: float = 7.0,
        optimize_ligand: bool = True,
        pdb_name: Optional[str] = None,
) -> Tuple[RDKitMol, RDKitMol]:
    """This prepares protein-ligand complexes for docking.
    Autodock Vina requires PDB files for proteins and ligands with
    sensible inputs. This function uses PDBFixer and RDKit to ensure
    that inputs are reasonable and ready for docking. Default values
    are given for convenience, but fixing PDB files is complicated and
    human judgement is required to produce protein structures suitable
    for docking. Always inspect the results carefully before trying to
    perform docking.

    Parameters
    ----------
    protein: str
      Filename for protein PDB file or a PDBID.
    ligand: str
      Either a filename for a ligand PDB file or a SMILES string.
    replace_nonstandard_residues: bool (default True)
      Replace nonstandard residues with standard residues.
    remove_heterogens: bool (default True)
      Removes residues that are not standard amino acids or nucleotides.
    remove_water: bool (default True)
      Remove water molecules.
    add_hydrogens: bool (default True)
      Add missing hydrogens at the protonation state given by `pH`.
    pH: float (default 7.0)
      Most common form of each residue at given `pH` value is used.
    optimize_ligand: bool (default True)
      If True, optimize ligand with RDKit. Required for SMILES inputs.
    pdb_name: Optional[str]
      If given, write sanitized protein and ligand to files called
      "pdb_name.pdb" and "ligand_pdb_name.pdb"

    Returns
    -------
    Tuple[RDKitMol, RDKitMol]
      Tuple of `protein_molecule, ligand_molecule` with 3D information.
    Note
    ----
    This function requires RDKit and OpenMM to be installed.
    Read more about PDBFixer here: https://github.com/openmm/pdbfixer.

    Examples
    --------
    >>> p, m = prepare_inputs('3cyx', 'CCC')
    >> p.GetNumAtoms()
    >> m.GetNumAtoms()
    >>> p, m = prepare_inputs('3cyx', 'CCC', remove_heterogens=False)
    >> p.GetNumAtoms()

    """
    try:
        from rdkit import Chem
        from rdkit.Chem import AllChem
        from pdbfixer import PDBFixer
        from simtk.openmm.app import PDBFile
    except ModuleNotFoundError:
        raise ImportError("This function requires RDKit and OpenMM to be installed.")

    if protein.endswith(".pdb"):
        fixer = PDBFixer(protein)
    else:
        fixer = PDBFixer(url="https://files.rcsb.org/download/%s.pdb" % (protein))

    if ligand.endswith(".pdb"):
        m = Chem.MolFromPDBFile(ligand)
    else:
        m = Chem.MolFromSmiles(ligand, sanitize=True)

    # Apply common fixes to PDB files
    if replace_nonstandard_residues:
        fixer.findMissingResidues()
        fixer.findNonstandardResidues()
        fixer.replaceNonstandardResidues()
    if remove_heterogens and not remove_water:
        fixer.removeHeterogens(True)
    if remove_heterogens and remove_water:
        fixer.removeHeterogens(False)
    if add_hydrogens:
        fixer.addMissingHydrogens(pH)

    PDBFile.writeFile(fixer.topology, fixer.positions, open("tmp.pdb", "w"))
    p = Chem.MolFromPDBFile("tmp.pdb", sanitize=True)
    os.remove("tmp.pdb")

    # Optimize ligand
    if optimize_ligand:
        m = Chem.AddHs(m)  # need hydrogens for optimization
        AllChem.EmbedMolecule(m)
        AllChem.MMFFOptimizeMolecule(m)

    if pdb_name:
        Chem.rdmolfiles.MolToPDBFile(p, "%s.pdb" % (pdb_name))
        Chem.rdmolfiles.MolToPDBFile(m, "ligand_%s.pdb" % (pdb_name))

    return (p, m)


class CoordinateBox(object):
    """A coordinate box that represents a block in space.
    Molecular complexes are typically represented with atoms as
    coordinate points. Each complex is naturally associated with a
    number of different box regions. For example, the bounding box is a
    box that contains all atoms in the molecular complex. A binding
    pocket box is a box that focuses in on a binding region of a protein
    to a ligand. A interface box is the region in which two proteins
    have a bulk interaction.
    The `CoordinateBox` class is designed to represent such regions of
    space. It consists of the coordinates of the box, and the collection
    of atoms that live in this box alongside their coordinates.
    """

    def __init__(
            self,
            x_range: Tuple[float, float],
            y_range: Tuple[float, float],
            z_range: Tuple[float, float],
    ):
        """Initialize this box.

        Parameters
        ----------
        x_range: Tuple[float, float]
          A tuple of `(x_min, x_max)` with max and min x-coordinates.
        y_range: Tuple[float, float]
          A tuple of `(y_min, y_max)` with max and min y-coordinates.
        z_range: Tuple[float, float]
          A tuple of `(z_min, z_max)` with max and min z-coordinates.

        Raises
        ------
        `ValueError` if this interval is malformed

        """
        if not isinstance(x_range, tuple) or not len(x_range) == 2:
            raise ValueError("x_range must be a tuple of length 2")
        else:
            x_min, x_max = x_range
            if not x_min <= x_max:
                raise ValueError("x minimum must be <= x maximum")
        if not isinstance(y_range, tuple) or not len(y_range) == 2:
            raise ValueError("y_range must be a tuple of length 2")
        else:
            y_min, y_max = y_range
            if not y_min <= y_max:
                raise ValueError("y minimum must be <= y maximum")
        if not isinstance(z_range, tuple) or not len(z_range) == 2:
            raise ValueError("z_range must be a tuple of length 2")
        else:
            z_min, z_max = z_range
            if not z_min <= z_max:
                raise ValueError("z minimum must be <= z maximum")
        self.x_range = x_range
        self.y_range = y_range
        self.z_range = z_range

    def __repr__(self):
        """Create a string representation of this box"""
        x_str = str(self.x_range)
        y_str = str(self.y_range)
        z_str = str(self.z_range)
        return "Box[x_bounds=%s, y_bounds=%s, z_bounds=%s]" % (x_str, y_str, z_str)

    def __str__(self):
        """Create a string representation of this box."""
        return self.__repr__()

    def __contains__(self, point: Sequence[float]) -> bool:
        """Check whether a point is in this box.

        Parameters
        ----------
        point: Sequence[float]
          3-tuple or list of length 3 or np.ndarray of shape `(3,)`.
          The `(x, y, z)` coordinates of a point in space.

        Returns
        -------
        bool
          `True` if `other` is contained in this box.

        """
        (x_min, x_max) = self.x_range
        (y_min, y_max) = self.y_range
        (z_min, z_max) = self.z_range
        x_cont = x_min <= point[0] and point[0] <= x_max
        y_cont = y_min <= point[1] and point[1] <= y_max
        z_cont = z_min <= point[2] and point[2] <= z_max
        return x_cont and y_cont and z_cont

    # FIXME: Argument 1 of "__eq__" is incompatible with supertype "object"
    def __eq__(self, other: "CoordinateBox") -> bool:  # type: ignore
        """Compare two boxes to see if they're equal.

        Parameters
        ----------
        other: CoordinateBox
          Compare this coordinate box to the other one.

        Returns
        -------
        bool
          That's `True` if all bounds match.

        Raises
        ------
        `ValueError` if attempting to compare to something that isn't a
        `CoordinateBox`.

        """
        if not isinstance(other, CoordinateBox):
            raise ValueError("Can only compare to another box.")
        return (
                self.x_range == other.x_range
                and self.y_range == other.y_range
                and self.z_range == other.z_range
        )

    def __hash__(self) -> int:
        """Implement hashing function for this box.
        Uses the default `hash` on `self.x_range, self.y_range,
        self.z_range`.

        Returns
        -------
        int
          Unique integer

        """
        return hash((self.x_range, self.y_range, self.z_range))

    def center(self) -> Tuple[float, float, float]:
        """Computes the center of this box.

        Returns
        -------
        Tuple[float, float, float]
          `(x, y, z)` the coordinates of the center of the box.

        Examples
        --------
        >>> box = CoordinateBox((0, 1), (0, 1), (0, 1))
        >>> box.center()
        (0.5, 0.5, 0.5)

        """
        x_min, x_max = self.x_range
        y_min, y_max = self.y_range
        z_min, z_max = self.z_range
        return (
            x_min + (x_max - x_min) / 2,
            y_min + (y_max - y_min) / 2,
            z_min + (z_max - z_min) / 2,
        )

    def volume(self) -> float:
        """Computes and returns the volume of this box.

        Returns
        -------
        float
          The volume of this box. Can be 0 if box is empty

        Examples
        --------
        >>> box = CoordinateBox((0, 1), (0, 1), (0, 1))
        >>> box.volume()
        1

        """
        x_min, x_max = self.x_range
        y_min, y_max = self.y_range
        z_min, z_max = self.z_range
        return (x_max - x_min) * (y_max - y_min) * (z_max - z_min)

    def contains(self, other: "CoordinateBox") -> bool:
        """Test whether this box contains another.
        This method checks whether `other` is contained in this box.

        Parameters
        ----------
        other: CoordinateBox
          The box to check is contained in this box.

        Returns
        -------
        bool
          `True` if `other` is contained in this box.

        Raises
        ------
        `ValueError` if `not isinstance(other, CoordinateBox)`.

        """
        if not isinstance(other, CoordinateBox):
            raise ValueError("other must be a CoordinateBox")
        other_x_min, other_x_max = other.x_range
        other_y_min, other_y_max = other.y_range
        other_z_min, other_z_max = other.z_range
        self_x_min, self_x_max = self.x_range
        self_y_min, self_y_max = self.y_range
        self_z_min, self_z_max = self.z_range
        return (
                self_x_min <= other_x_min
                and other_x_max <= self_x_max
                and self_y_min <= other_y_min
                and other_y_max <= self_y_max
                and self_z_min <= other_z_min
                and other_z_max <= self_z_max
        )
